/**
 * Demo 10 - Particles.cpp
 * @author Brian Sowers
 * Handles a very simple particle system.
 */
#include <cstdlib>
#include <iostream>
#include <string>
#include <ctime>
#include <list>
#include <gl/glut.h>
#include <gl/glaux.h>
#include <gl/glu.h>
using namespace std;

//Structure to store a single particle
struct Particle
{
	float pos[3];
	float vel[3];
	float color[3];
	int life;
};

void init();
void reshape(int w, int h);
void idle();
void display();

void mouse(int button, int state, int x, int y);
void mouseMotion(int x, int y);
void keyboard(unsigned char key, int x, int y);

int loadTexture(const string &fileName);

void drawParticles();
void updateParticles();
void generateExplosion(int numParticles);
void updateFire();

//List of all the particles in the scene
list<Particle> particles;
bool fireOn = false;

/**
 * Entry point of the program.
 * @param argc number of command-line parameters
 * @param argv the command-line parameters
 */
int main(int argc, char *argv[])
{
	//Initializes glut
	glutInit(&argc, argv);

	//Sets the size of the window that will be displayed
	glutInitWindowSize(800, 600);

	//Sets the window's initial position
	glutInitWindowPosition(50, 50);

	//Sets the display mode
	//GLUT_RGBA says that we will be dealing with 32 bit color, where each color has a red, green, blue,
	//and alpha component.
	//GLUT_DOUBLE says that we will have a double buffered application.  In a double buffered application,
	//whenever something is drawn, it is drawn to a 'back buffer' which is not immediately displayed.
	//Only when drawing finishes is the contents of the back buffer displayed to the screen.
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);

	//Creates the window with a given title bar
	glutCreateWindow("OpenGL Guide");
	init();

	//Sets the reshape function to a function we created.  This function is called automatically when the window is resized.
	glutReshapeFunc(reshape);

	//Sets the display function.  This function is called whenever glut determines that something
	//should be drawn to the screen.
	glutDisplayFunc(display);

	//Sets the idle function.  This function is called whenever glut is not doing anything else.
	glutIdleFunc(idle);
	
	//Sets the mouse function.  This function is called when a mouse button is pressed.
	glutMouseFunc(mouse);

	//Sets the mouse motion function.  This function is called when the mouse moves.
	glutMotionFunc(mouseMotion);

	//Sets the keyboard function.  This function is called when a key is typed.
	glutKeyboardFunc(keyboard);

	//Enters the glut main loop.  This loop executes glut code repeatedly, calling the functions
	//that we set previously with glutReshapeFunc, glutDisplayFunc, etc at the appropriate times.
	//It is important to note that this function never ends (there is an infinite loop), so code
	//after this function can not be executed.
	glutMainLoop();
	return 0;
}

/**
 * Performs necessary initialization.
 */
void init()
{
	glClearColor(0,0,0,0);

	//Enables depth testing.  When depth testing is turned on, OpenGL will
	//properly sort objects drawn behind one-another.  When depth testing
	//is turned off, this sorting will not occur - objects will be drawn
	//on top of one-another despite where they exist in the scene.
	glEnable(GL_DEPTH_TEST);
}

/**
 * This function is called whenever glut is not doing something else.
 */
void idle()
{
	updateParticles();
	if (fireOn)
	{
		updateFire();
	}

	//Forces the display function to be called at the next available moment
	glutPostRedisplay();
}

/**
 * Called automatically by glut when the window is resized.
 * @param w the new width of the window
 * @param h the new height of the window
 */
void reshape(int w, int h)
{
	//Sets the viewport into which OpenGL will draw
	glViewport(0, 0, w, h);

	//Sets up a 3D perspective projection.
	//There are two primary types of projection matrices: orthographic and perspective.
	//In orthographic, objects appear 'flat' - that is, they do not shrink or expand
	//in relation to their position from the viewer.
	//Perspective is more realistic.  It makes objects appear more realistic, causing
	//them to resize depending on te position from the viewer.
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	//Here we use a utility function to setup a perspective projection.
	//The first parameter is the field of view.
	//The second parameter is the aspect ratio
	//The third is the near clip plane - how close objects will be before they are considered too close to draw
	//The fourth is the far clip plane - how far away objects can be before OpenGL stops drawing them
	gluPerspective(60, (float)w/(float)h, 1.0, 1000.0);
}

/**
 * Called automatically by glut when the screen needs redrawing.  We can also
 * force this function to be called with glutPostRedisplay
 */
void display()
{
	//Clears the screen.
	//GL_COLOR_BUFFER_BIT tells OpenGL to clear the color buffer
	//GL_DEPTH_BUFFER_BIT tells OpenGL to clear the depth buffer
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//Restores the basic modelview matrix.  The modelview matrix determines how an object
	//will be moved/rotated/scaled before being rendered.
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	//Sets up the camera.  This camera has 3 basic vectors associated with it - a position, a lookAt,
	//and an 'up' fector.
	//The position (the first 3 parameters) represents where the camera is located in 3d space
	//The lookAt represents where the camera is looking
	//Up represents the direction 'up' is in relation to the camera.  Often, (0,1,0) suffices.
	gluLookAt(0,150,150, 0,0,0, 0,1,0);

	drawParticles();

	//Swaps buffers so that the contents of the back buffer are displayed.
	glutSwapBuffers();
}

/**
 * Called when a mouse button's state changes.
 * @param button the button that changed
 * @param state teh button's new state
 * @param x the x position of the mouse when the state change happened
 * @param y the y position of the mouse when the state change happened
 */
void mouse(int button, int state, int x, int y)
{
	
}

/**
 * Called when the mouse moves.
 * @param x new x position of the mouse
 * @param y new y position of the mouse
 */
void mouseMotion(int x, int y)
{
} 

/**
 * Called when the user types a key.
 * @param key key the user typed
 * @param x x position of the mouse when the key was typed
 * @param y y position of the mouse when the key was typed
 */
void keyboard(unsigned char key, int x, int y)
{
	if (key == 'e')
	{
		generateExplosion(100);
	}
	if (key == 'f')
	{
		fireOn = !fireOn;
	}
}

/**
 * Loads a texture and returns the texture ID which can be used by OpenGL
 * @param fileName name of the texture to open
 * @return texture ID
 */
int loadTexture(const string &fileName)
{
	AUX_RGBImageRec *image = auxDIBImageLoad(fileName.c_str());

	unsigned int textureID;

	//Generates a valid texture ID
	glGenTextures(1, &textureID);

	//Binds our newly generated texture so that all further texture operations apply to it
	glBindTexture(GL_TEXTURE_2D, textureID);

	//Loads our image data into texture memory
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, image->sizeX, image->sizeY, 0, GL_RGB, GL_UNSIGNED_BYTE, image->data);

	//Sets up the filtering methods, which determine how the image stretches/scales
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);

	//Frees the image data.  We can do this now since there is a copy in texture memory.
	free(image->data);
	free(image);

	return textureID;
}

/**
 * Updates a particle's position and lifetime, killing particles when necessary.
 */
void updateParticles()
{
	list<Particle>::iterator i = particles.begin();
	while (i != particles.end())
	{
		if ( (*i).life <= 0)
		{
			i = particles.erase(i);
		}
		else
		{
			(*i).life--;
			(*i).pos[0] = (*i).pos[0] + (*i).vel[0];
			(*i).pos[1] = (*i).pos[1] + (*i).vel[1];
			(*i).pos[2] = (*i).pos[2] + (*i).vel[2];

			i++;
		}
	}
}

/**
 * Draws all the particles as spheres.
 */
void drawParticles()
{
	list<Particle>::iterator i = particles.begin();
	while (i != particles.end())
	{
		Particle p = (*i);

		glPushMatrix();
			glTranslatef(p.pos[0], p.pos[1], p.pos[2]);
			glColor3fv(p.color);

			glutSolidSphere(1.0, 4.0, 4.0);

		glPopMatrix();

		i++;
	}

}

/**
 * Generates a random explosion of multicolored particles.
 * @param numParticles number of particles involved in the explosion
 */
void generateExplosion(int numParticles)
{
	srand( (unsigned int)time(0) );
	for (int particle = 0; particle < numParticles; particle++)
	{
		Particle p;
		p.pos[0] = p.pos[1] = p.pos[2] = 0;
		p.vel[0] = rand()/(float)RAND_MAX - 0.5;
		p.vel[1] = rand()/(float)RAND_MAX - 0.5;
		p.vel[2] = rand()/(float)RAND_MAX - 0.5;
		p.color[0] = rand()/(float)RAND_MAX;
		p.color[1] = rand()/(float)RAND_MAX;
		p.color[2] = rand()/(float)RAND_MAX;
		p.life = 200;

		particles.push_back(p);
	}
}

/**
 * Updates a particle fire (if it's turned on).
 * If the fire is turned on, every frame a few new particles are created.
 */
void updateFire()
{
	for (int particle = 0; particle < 5; particle++)
	{
		Particle p;
		p.pos[0] = rand()/(float)RAND_MAX * 5.0 - 2.5;
		p.pos[1] = rand()/(float)RAND_MAX;
		p.pos[2] = 0;
		p.vel[0] = 0;
		p.vel[1] = rand()/(float)RAND_MAX;
		p.vel[2] = 0;
		p.color[0] = rand()/(float)RAND_MAX;
		p.color[1] = 0;
		p.color[2] = 0;
		p.life = 100;

		particles.push_back(p);
	}
}
