/**
 * Demo 2 - Basic Input.cpp
 * @author Brian Sowers
 * This demo illustrates how to capture mouse and keyboard input.
 */
#include <cstdlib>
#include <iostream>
#include <gl/glut.h>
using namespace std;

void init();
void reshape(int w, int h);
void idle();
void display();

void mouse(int button, int state, int x, int y);
void mouseMotion(int x, int y);
void keyboard(unsigned char key, int x, int y);

/**
 * Entry point of the program.
 * @param argc number of command-line parameters
 * @param argv the command-line parameters
 */
int main(int argc, char *argv[])
{
	//Initializes glut
	glutInit(&argc, argv);

	//Sets the size of the window that will be displayed
	glutInitWindowSize(800, 600);

	//Sets the window's initial position
	glutInitWindowPosition(50, 50);

	//Sets the display mode
	//GLUT_RGBA says that we will be dealing with 32 bit color, where each color has a red, green, blue,
	//and alpha component.
	//GLUT_DOUBLE says that we will have a double buffered application.  In a double buffered application,
	//whenever something is drawn, it is drawn to a 'back buffer' which is not immediately displayed.
	//Only when drawing finishes is the contents of the back buffer displayed to the screen.
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);

	//Creates the window with a given title bar
	glutCreateWindow("OpenGL Guide");
	init();

	//Sets the reshape function to a function we created.  This function is called automatically when the window is resized.
	glutReshapeFunc(reshape);

	//Sets the display function.  This function is called whenever glut determines that something
	//should be drawn to the screen.
	glutDisplayFunc(display);

	//Sets the idle function.  This function is called whenever glut is not doing anything else.
	glutIdleFunc(idle);
	
	//Sets the mouse function.  This function is called when a mouse button is pressed.
	glutMouseFunc(mouse);

	//Sets the mouse motion function.  This function is called when the mouse moves.
	glutMotionFunc(mouseMotion);

	//Sets the keyboard function.  This function is called when a key is typed.
	glutKeyboardFunc(keyboard);

	//Enters the glut main loop.  This loop executes glut code repeatedly, calling the functions
	//that we set previously with glutReshapeFunc, glutDisplayFunc, etc at the appropriate times.
	//It is important to note that this function never ends (there is an infinite loop), so code
	//after this function can not be executed.
	glutMainLoop();
	return 0;
}

/**
 * Performs necessary initialization.
 */
void init()
{
	glClearColor(0,0,0,0);

	//Enables depth testing.  When depth testing is turned on, OpenGL will
	//properly sort objects drawn behind one-another.  When depth testing
	//is turned off, this sorting will not occur - objects will be drawn
	//on top of one-another despite where they exist in the scene.
	glEnable(GL_DEPTH_TEST);
}

/**
 * This function is called whenever glut is not doing something else.
 */
void idle()
{
	//Forces the display function to be called at the next available moment
	glutPostRedisplay();
}

/**
 * Called automatically by glut when the window is resized.
 * @param w the new width of the window
 * @param h the new height of the window
 */
void reshape(int w, int h)
{
	//Sets the viewport into which OpenGL will draw
	glViewport(0, 0, w, h);
}

/**
 * Called automatically by glut when the screen needs redrawing.  We can also
 * force this function to be called with glutPostRedisplay
 */
void display()
{
	//Clears the screen.
	//GL_COLOR_BUFFER_BIT tells OpenGL to clear the color buffer
	//GL_DEPTH_BUFFER_BIT tells OpenGL to clear the depth buffer
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//Swaps buffers so that the contents of the back buffer are displayed.
	glutSwapBuffers();
}

/**
 * Called when a mouse button's state changes.
 * @param button the button that changed
 * @param state teh button's new state
 * @param x the x position of the mouse when the state change happened
 * @param y the y position of the mouse when the state change happened
 */
void mouse(int button, int state, int x, int y)
{
	if (button == GLUT_LEFT_BUTTON)
	{
		if (state == GLUT_UP)
		{
			cout << "Left mouse button released." << endl;
		}
		if (state == GLUT_DOWN)
		{
			cout << "Left mouse button pressed." << endl;
		}
	}
	else if (button == GLUT_RIGHT_BUTTON)
	{
		if (state == GLUT_UP)
		{
			cout << "Right mouse button released." << endl;
		}
		if (state == GLUT_DOWN)
		{
			cout << "Right mouse button pressed." << endl;
		}
	}
}

/**
 * Called when the mouse moves.
 * @param x new x position of the mouse
 * @param y new y position of the mouse
 */
void mouseMotion(int x, int y)
{
}

/**
 * Called when the user types a key.
 * @param key key the user typed
 * @param x x position of the mouse when the key was typed
 * @param y y position of the mouse when the key was typed
 */
void keyboard(unsigned char key, int x, int y)
{
	cout << "Key typed: " << key << endl;
}
