/**
 * Demo 9 - Terrain.cpp
 * @author Brian Sowers
 * Renders a very simple, randomly generated terrain which is also texture mapped (but not lit).
 */
#include <cstdlib>
#include <iostream>
#include <string>
#include <ctime>
#include <cmath>
#include <gl/glut.h>
#include <gl/glaux.h>
#include <gl/glu.h>
using namespace std;

void init();
void reshape(int w, int h);
void idle();
void display();

void mouse(int button, int state, int x, int y);
void mouseMotion(int x, int y);
void keyboard(unsigned char key, int x, int y);

void generateRandomTerrain();
void drawTerrain();
void loadTerrain(const string &fileName);

int loadTexture(const string &fileName);

int texture;

const int TERRAIN_X = 256;
const int TERRAIN_Y = 256;
float heightMap[TERRAIN_X][TERRAIN_Y];
float maxHeight = 0;
float rotateX = 0.0, rotateY = 0.0;

/**
 * Entry point of the program.
 * @param argc number of command-line parameters
 * @param argv the command-line parameters
 */
int main(int argc, char *argv[])
{
	//Initializes glut
	glutInit(&argc, argv);

	//Sets the size of the window that will be displayed
	glutInitWindowSize(800, 600);

	//Sets the window's initial position
	glutInitWindowPosition(50, 50);

	//Sets the display mode
	//GLUT_RGBA says that we will be dealing with 32 bit color, where each color has a red, green, blue,
	//and alpha component.
	//GLUT_DOUBLE says that we will have a double buffered application.  In a double buffered application,
	//whenever something is drawn, it is drawn to a 'back buffer' which is not immediately displayed.
	//Only when drawing finishes is the contents of the back buffer displayed to the screen.
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);

	//Creates the window with a given title bar
	glutCreateWindow("OpenGL Guide");
	init();

	//Sets the reshape function to a function we created.  This function is called automatically when the window is resized.
	glutReshapeFunc(reshape);

	//Sets the display function.  This function is called whenever glut determines that something
	//should be drawn to the screen.
	glutDisplayFunc(display);

	//Sets the idle function.  This function is called whenever glut is not doing anything else.
	glutIdleFunc(idle);
	
	//Sets the mouse function.  This function is called when a mouse button is pressed.
	glutMouseFunc(mouse);

	//Sets the mouse motion function.  This function is called when the mouse moves.
	glutMotionFunc(mouseMotion);

	//Sets the keyboard function.  This function is called when a key is typed.
	glutKeyboardFunc(keyboard);

	//Enters the glut main loop.  This loop executes glut code repeatedly, calling the functions
	//that we set previously with glutReshapeFunc, glutDisplayFunc, etc at the appropriate times.
	//It is important to note that this function never ends (there is an infinite loop), so code
	//after this function can not be executed.
	glutMainLoop();
	return 0;
}

/**
 * Performs necessary initialization.
 */
void init()
{
	glClearColor(0,0,0,0);

	texture = loadTexture("worldtexture.bmp");

	loadTerrain("heightmap.bmp");

	//Enables depth testing.  When depth testing is turned on, OpenGL will
	//properly sort objects drawn behind one-another.  When depth testing
	//is turned off, this sorting will not occur - objects will be drawn
	//on top of one-another despite where they exist in the scene.
	glEnable(GL_DEPTH_TEST);
}

/**
 * This function is called whenever glut is not doing something else.
 */
void idle()
{
	//Forces the display function to be called at the next available moment
	glutPostRedisplay();
}

/**
 * Called automatically by glut when the window is resized.
 * @param w the new width of the window
 * @param h the new height of the window
 */
void reshape(int w, int h)
{
	//Sets the viewport into which OpenGL will draw
	glViewport(0, 0, w, h);

	//Sets up a 3D perspective projection.
	//There are two primary types of projection matrices: orthographic and perspective.
	//In orthographic, objects appear 'flat' - that is, they do not shrink or expand
	//in relation to their position from the viewer.
	//Perspective is more realistic.  It makes objects appear more realistic, causing
	//them to resize depending on te position from the viewer.
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	//Here we use a utility function to setup a perspective projection.
	//The first parameter is the field of view.
	//The second parameter is the aspect ratio
	//The third is the near clip plane - how close objects will be before they are considered too close to draw
	//The fourth is the far clip plane - how far away objects can be before OpenGL stops drawing them
	gluPerspective(60, (float)w/(float)h, 1.0, 1000.0);
}

/**
 * Called automatically by glut when the screen needs redrawing.  We can also
 * force this function to be called with glutPostRedisplay
 */
void display()
{
	//Clears the screen.
	//GL_COLOR_BUFFER_BIT tells OpenGL to clear the color buffer
	//GL_DEPTH_BUFFER_BIT tells OpenGL to clear the depth buffer
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//Restores the basic modelview matrix.  The modelview matrix determines how an object
	//will be moved/rotated/scaled before being rendered.
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	//Sets up the camera.  This camera has 3 basic vectors associated with it - a position, a lookAt,
	//and an 'up' fector.
	//The position (the first 3 parameters) represents where the camera is located in 3d space
	//The lookAt represents where the camera is looking
	//Up represents the direction 'up' is in relation to the camera.  Often, (0,1,0) suffices.
	gluLookAt(0,150,150, 0,0,0, 0,1,0);

	//Enables texture mapping
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, texture);

	glRotatef(rotateX, 1, 0, 0);
	glRotatef(rotateY, 0, 1, 0);

	//Renders the terrain based on the heightmap
	drawTerrain();

	//Disable our lights
	glDisable(GL_LIGHT0);
	glDisable(GL_LIGHTING);

	//Disables texturing just in case further calls should not use it.
	glDisable(GL_TEXTURE_2D);
	
	//Swaps buffers so that the contents of the back buffer are displayed.
	glutSwapBuffers();
}

/**
 * Called when a mouse button's state changes.
 * @param button the button that changed
 * @param state teh button's new state
 * @param x the x position of the mouse when the state change happened
 * @param y the y position of the mouse when the state change happened
 */
void mouse(int button, int state, int x, int y)
{
	
}

/**
 * Called when the mouse moves.
 * @param x new x position of the mouse
 * @param y new y position of the mouse
 */
void mouseMotion(int x, int y)
{
	static int oldX = x;
	static int oldY = y;

	rotateY = (x - oldX)/2.0;
	rotateX = (y - oldY)/2.0;
} 

/**
 * Called when the user types a key.
 * @param key key the user typed
 * @param x x position of the mouse when the key was typed
 * @param y y position of the mouse when the key was typed
 */
void keyboard(unsigned char key, int x, int y)
{
}

/**
 * Loads a texture and returns the texture ID which can be used by OpenGL
 * @param fileName name of the texture to open
 * @return texture ID
 */
int loadTexture(const string &fileName)
{
	AUX_RGBImageRec *image = auxDIBImageLoad(fileName.c_str());

	unsigned int textureID;

	//Generates a valid texture ID
	glGenTextures(1, &textureID);

	//Binds our newly generated texture so that all further texture operations apply to it
	glBindTexture(GL_TEXTURE_2D, textureID);

	//Loads our image data into texture memory
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, image->sizeX, image->sizeY, 0, GL_RGB, GL_UNSIGNED_BYTE, image->data);

	//Sets up the filtering methods, which determine how the image stretches/scales
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);

	//Frees the image data.  We can do this now since there is a copy in texture memory.
	free(image->data);
	free(image);

	return textureID;
}

/**
 * Generates a completely random terrain.  This has some ugly results.
 * A better solution would be to use a greyscale heightmap or a Perlin noise function
 */
void generateRandomTerrain()
{
	srand( (unsigned int)time(0) );
	for (int x = 0; x < TERRAIN_X; x++)
	{
		for (int y = 0; y < TERRAIN_Y; y++)
		{
			heightMap[x][y] = rand()%15;
		}
	}
}

/**
 * Loads a heightmap from a 256x256 RGB bitmap file.
 * @param fileName name of the file to use for the height map
 */
void loadTerrain(const string &fileName)
{
	AUX_RGBImageRec *image = auxDIBImageLoad(fileName.c_str());
	int index = 0;
	
	for (int y = 0; y < TERRAIN_Y; y++)
	{
		for (int x = 0; x < TERRAIN_X; x++)
		{
			heightMap[x][y] = image->data[index] / 5.0;
			if (heightMap[x][y] > maxHeight) maxHeight = heightMap[x][y];

			index += 3;
		}
	}

	free(image->data);
	free(image);
}

/**
 * Draws the terrain
 */
void drawTerrain()
{
	glColor3f(1,1,1);
	for (int x = 0; x < TERRAIN_X - 1; x++)
	{
		glBegin(GL_TRIANGLE_STRIP);
		for (int y = 0; y < TERRAIN_Y; y++)
			{
				float shade = heightMap[x][y] / maxHeight + 0.2;
				glColor3f(shade, shade, shade);
				glTexCoord2f((float)x/TERRAIN_X, (float)y/TERRAIN_Y);
				glVertex3f(x - TERRAIN_X/2, heightMap[x][y], y - TERRAIN_Y/2);

				shade = heightMap[x+1][y] / maxHeight + 0.2;
				glColor3f(shade, shade, shade);
				glTexCoord2f((float)(x+1)/TERRAIN_X, (float)y/TERRAIN_Y);
				glVertex3f(x+1 - TERRAIN_X/2, heightMap[x+1][y], y - TERRAIN_Y/2);
			}
		glEnd();
	}
}
